(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_datanode_ctrl.ml 239 2010-06-23 16:49:03Z gerd $ *)

module Rpcapi_clnt = Pfs_rpcapi_clnt
module Rpcapi_srv = Pfs_rpcapi_srv
module Rpcapi_aux = Pfs_rpcapi_aux

module Dnctrl_proxy = Rpcapi_clnt.Make'Datanode_ctrl(Rpc_proxy.ManagedClient)

type client = Rpc_proxy.ManagedClient.mclient

type ctrl =
    { clients : (int, Nn_datastores.datastore * client) Hashtbl.t;
      (** maps datastore ID to client *)

    }

module Ctrl_var = 
  Netplex_cenv.Make_var_type(struct type t = ctrl end)

let ctrlname = "current_datanode_ctrl"

let get_ctrl() =
  Ctrl_var.get ctrlname

let post_start_hook() =
  Ctrl_var.set ctrlname
    { clients = Hashtbl.create 7 }


let create_client conn esys =
  let mclient_config =
    Rpc_proxy.ManagedClient.create_mclient_config
      ~programs:[ Rpcapi_clnt.Datanode_ctrl.V1._program ]
      ~msg_timeout:(60.0)
      ~msg_timeout_is_fatal:true
      ~initial_ping:true
      () in
  let mclient =
    Rpc_proxy.ManagedClient.create_mclient
      mclient_config
      conn
      esys in
  mclient


let live_datastores() =
  (** Returns the list of live datastores *)
  let ds_list =
    Nn_datastore_news.list ~wait:true () in
  List.filter
    (fun ds ->
       let identity = ds.Nn_datastores.identity in
       ds.Nn_datastores.enabled && Nn_datastore_news.is_alive identity
    )
    ds_list


let reset_new_datastores() =
  (** Create connections to newly added datastores, and send a reset to them *)
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let ctrl = get_ctrl() in
  let ds_list = live_datastores() in
  List.iter
    (fun ds ->
       let id = ds.Nn_datastores.id in
       if not (Hashtbl.mem ctrl.clients id) then (
	 let conn =
	   match ds.Nn_datastores.node with
	     | None -> assert false
	     | Some addr -> Plasma_util.connector_of_sockaddr addr in
	 let client = create_client conn esys in
	 Dnctrl_proxy.V1.reset_all_safetrans'async
	   client
	   ()
	   (fun get_reply ->
	      try let () = get_reply() in ()
	      with error ->
		Netlog.logf `Err
		  "reset_all_safetrans: Exception %s"
		  (Netexn.to_string error);
		Rpc_proxy.ManagedClient.enforce_unavailability client;
		Hashtbl.remove ctrl.clients id
	   );
	 Hashtbl.add ctrl.clients id (ds,client)
       )
    )
    ds_list
  

let disconnect id =
  (** Look the client up, and terminate the connection: *)
  let ctrl = get_ctrl() in
  try
    let _, client = Hashtbl.find ctrl.clients id in
    Rpc_proxy.ManagedClient.trigger_shutdown client (fun () -> ());
    Hashtbl.remove ctrl.clients id
  with
    | Not_found -> ()


let reset() =
  (** First terminate all connections: *)
  let ctrl = get_ctrl() in
  Hashtbl.iter
    (fun id (_,client) ->
       Rpc_proxy.ManagedClient.trigger_shutdown client (fun () -> ())
    )
    ctrl.clients;
  Hashtbl.clear ctrl.clients;

  (** Get new list of datastores, create new clients, and send the reset: *)
  reset_new_datastores()


let safetrans_e ~st_id ~st_tmo ~st_secret =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let ctrl = get_ctrl() in
  let ds_list = live_datastores() in
  reset_new_datastores();
  
  let l = ref [] in

  new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done !l)
    (Plasma_util.msync_engine
       ( List.map
	   (fun ds ->
	      let id = ds.Nn_datastores.id in
	      let _, client = Hashtbl.find ctrl.clients id in
	      new Uq_engines.map_engine
		~map_done:(fun _ -> 
			     Netlog.logf `Debug
			       "Datastore %s: created safetrans %Ld secret %Ld"
			       ds.Nn_datastores.identity st_id st_secret;
			     l := id :: !l; 
			     `Done())
		~map_error:(fun e ->
			      Netlog.logf `Err
				"safetrans: Exception %s"
				(Netexn.to_string e);
			      `Done ())
		(Plasma_util.rpc_engine
		   client
		   Dnctrl_proxy.V1.safetrans'async
		   (st_id, st_tmo, st_secret)
		)
	   )
	   ds_list
       )
       esys
    )


let cancel_safetrans_e ~st_id id_list =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let ctrl = get_ctrl() in
  Plasma_util.msync_engine
    ( List.map
	(fun id ->
	   try
	     let (ds,client) = Hashtbl.find ctrl.clients id in
	     new Uq_engines.map_engine
	       ~map_done:(fun _ -> 
			    Netlog.logf `Debug
			      "Datastore %s: cancelled safetrans %Ld"
			      ds.Nn_datastores.identity st_id;
			    `Done())
	       ~map_error:(fun e ->
			     Netlog.logf `Err
			       "cancel_safetrans: Exception %s"
			       (Netexn.to_string e);
			     `Done ())
	       (Plasma_util.rpc_engine
		  client
		  Dnctrl_proxy.V1.cancel_safetrans'async
		  st_id
	       )
	   with
	     | Not_found ->
		 new Uq_engines.epsilon_engine (`Done ()) esys
	)
	id_list
    )
    esys
